<?php

include_once(dirname(__FILE__).'/Page.php');

class ModuleMeasure extends AppModule
{
  function __construct($System)
  {
    parent::__construct($System);
    $this->Name = 'Measure';
    $this->Version = '1.0';
    $this->Creator = 'Chronos';
    $this->License = 'GNU/GPL';
    $this->Description = 'Measurement processing';
    $this->Dependencies = array();
  }

  function DoStart()
  {
    $this->System->RegisterPage('', 'PageMain');
  }
}

include_once('Point.php');

class Measure
{
  var $Data;
  var $ValueTypes;
  var $LevelReducing;
  var $Database;
  var $MaxLevel;
  var $ReferenceTime;
  var $Differential;
  var $DivisionCount;
  var $PeriodTolerance;

  function __construct(Database $Database)
  {
    $this->Id = 0;
    $this->ValueTypes = array('min', 'avg', 'max');
    $this->Database = &$Database;
    $this->LevelReducing = 5;
    $this->MaxLevel = 4;
    $this->ReferenceTime = 0;
    $this->Differential = 0;
    $this->DivisionCount = 500;
    $this->PeriodTolerance = 0.25; // 25%
  }

  function TimeSegment($Level)
  {
    return(pow($this->LevelReducing, $Level) * 60);
  }

  function StatTableName($Level)
  {
    if($Level == 0) return('data');
      else return('data_cache');
  }

  function AlignTime($Time, $TimeSegment)
  {
    return(round(($Time - $this->ReferenceTime) / $TimeSegment) * $TimeSegment +
      $this->ReferenceTime);
  }

  function Load($Id)
  {
    $Result = $this->Database->select('Measure', '*', '`Id`='.$Id);
    if($Result->num_rows > 0)
    {
      $this->Data = $Result->fetch_array();
      if($this->Data['Continuity'] == 0) $this->Data['ContinuityEnabled'] = 0;  // non continuous
        else $this->Data['ContinuityEnabled'] = 2;    // continuous graph
    } else die('Měřená veličina nenalezena');
  }

  function AddValue($Value)
  {
    $Time = time();

    $Result = $this->Database->select($this->Data['DataTable'], '*', '(`Measure`='.$this->Data['Id'].') AND '.
      '(`level`=0) ORDER BY `time` DESC LIMIT 2');
    if($Result->num_rows == 0) {
      // No measure value found. Simply insert new first value.
      $this->Database->insert($this->Data['DataTable'],
      array('min' => $Value, 'avg' => $Value, 'max' => $Value, 'level' => 0,
      'measure' => $this->Data['Id'], 'time' => TimeToMysqlDateTime($Time),
      'continuity' => 0));
    } else if($Result->num_rows == 1) {
      // One value exists. Add second value.
      $this->Database->insert($this->Data['DataTable'],
      array('min' => $Value, 'avg' => $Value, 'max' => $Value, 'level' => 0,
      'measure' => $this->Data['Id'], 'time' => TimeToMysqlDateTime($Time),
      'continuity' => 1));
    } else {
      // Two values already exist in measure table
      $LastValue = $Result->fetch_array();
      $NextToLastValue = $Result->fetch_array();
      if(($Time - MysqlDateTimeToTime($LastValue['time'])) < (1 - $this->PeriodTolerance) * $this->Data['Period'])
      {
        // New value added too quickly. Need to wait for next measure period.
      }
      else
      {
        // We are near defined period and can add new value.
        if(($Time - MysqlDateTimeToTime($LastValue['time'])) < (1 + $this->PeriodTolerance) * $this->Data['Period']) {
          // New value added near defined measure period inside tolerance. Keep continuity.
          $Continuity = 1;
        } else {
          // New value added too late after defined measure period. Stop
          // continuity and let gap to be added.
          $Continuity = 0;
        }
        if(($LastValue['avg'] == $NextToLastValue['avg']) and ($LastValue['avg'] == $Value) and
          ($LastValue['continuity'] == 1) and ($Continuity == 1))
        {
          // New value is same as last value and continuity mode is enabled and
          // continuity flag is present. Just shift forward time for last value.
          $this->Database->update($this->Data['DataTable'], '(`time`="'.$LastValue['time'].'") AND '.
          '(`level`=0) AND (`measure`='.$this->Data['Id'].')', array('time' => TimeToMysqlDateTime($Time)));
        } else
        {
          // Last value is different or not with continuity flag. Need to add new value.
          $this->Database->insert($this->Data['DataTable'], array('min' => $Value,
            'avg' => $Value, 'max' => $Value, 'level' => 0, 'measure' => $this->Data['Id'], 'time' => TimeToMysqlDateTime($Time),
            'continuity' => $Continuity));
        }
      }
    }

    // Update higher levels
    for($Level = 1; $Level <= $this->MaxLevel; $Level++)
    {
      $TimeSegment = $this->TimeSegment($Level);
      $EndTime = $this->AlignTime($Time, $TimeSegment);
      //if($EndTime < $Time) $EndTime = $EndTime + $TimeSegment;
      $StartTime = $EndTime - $TimeSegment;

      // Load values in time range
      $Values = array();
      $Result = $this->Database->select($this->Data['DataTable'], '*', '(`time` > "'.
        TimeToMysqlDateTime($StartTime).'") AND '.
        '(`time` < "'.TimeToMysqlDateTime($EndTime).'") AND '.
        '(`measure`='.$this->Data['Id'].') AND (`level`='.($Level - 1).') ORDER BY `time`');
      while($Row = $Result->fetch_array())
      {
        $Row['time'] = MysqlDateTimeToTime($Row['time']);
        $Values[] = $Row;
      }
      //array_pop($Values);

      // Load subsidary values
      $Values = array_merge(
        $this->LoadLeftSideValue($Level - 1, $StartTime),
        $Values,
        $this->LoadRightSideValue($Level - 1, $EndTime)
      );

      $Point = $this->ComputeOneValue($StartTime, $EndTime, $Values, $Level);

      $this->Database->delete($this->Data['DataTable'], '(`time` > "'.TimeToMysqlDateTime($StartTime).'") AND
        (`time` < "'.TimeToMysqlDateTime($EndTime).'") AND (`measure`='.$this->Data['Id'].') '.
        'AND (`level`='.$Level.')');
      $Continuity = $Values[1]['continuity'];
      $this->Database->insert($this->Data['DataTable'], array('level' => $Level,
        'measure' => $this->Data['Id'], 'min' => $Point['min'],
        'avg' => $Point['avg'], 'max' => $Point['max'], 'continuity' => $Continuity,
        'time' => TimeToMysqlDateTime($StartTime + ($EndTime - $StartTime) / 2)));
    }
  }

  /* Compute one value for upper time level from multiple values */
  function ComputeOneValue($LeftTime, $RightTime, $Values, $Level)
  {
    $NewValue = array('min' => +1000000000000000000, 'avg' => 0, 'max' => -1000000000000000000);

    // Trim outside parts
    foreach($this->ValueTypes as $ValueType)
    {
      $Values[0][$ValueType] = Interpolation(
        NewPoint($Values[0]['time'], $Values[0][$ValueType]),
        NewPoint($Values[1]['time'], $Values[1][$ValueType]), $LeftTime);
    }
    $Values[0]['time'] = $LeftTime;
    foreach($this->ValueTypes as $ValueType)
    {
      $Values[count($Values) - 1][$ValueType] = Interpolation(
        NewPoint($Values[count($Values) - 2]['time'], $Values[count($Values) - 2][$ValueType]),
        NewPoint($Values[count($Values) - 1]['time'], $Values[count($Values) - 1][$ValueType]),
        $RightTime);
    }
    $Values[count($Values) - 1]['time'] = $RightTime;

    // Perform computation
    foreach($this->ValueTypes as $ValueType)
    {
      // Compute new value
      for($I = 0; $I < (count($Values) - 1); $I++)
      {
        if($ValueType == 'avg')
        {
          if($Values[$I + 1]['continuity'] == $this->Data['ContinuityEnabled']);
            else if($this->Differential == 0)
            {
              $NewValue[$ValueType] = $NewValue[$ValueType] + ($Values[$I + 1]['time'] - $Values[$I]['time']) *
                (($Values[$I + 1][$ValueType] - $Values[$I][$ValueType]) / 2 + $Values[$I][$ValueType]);
            } else {
              $NewValue[$ValueType] = $NewValue[$ValueType] + ($Values[$I + 1]['time'] - $Values[$I]['time']) *
               (($Values[$I + 1][$ValueType] - $Values[$I][$ValueType]) / 2);
            }
        }
        else if($ValueType == 'max')
        {
          if($Values[$I + 1]['continuity'] == $this->Data['ContinuityEnabled'])
          {
            if(0 > $NewValue[$ValueType]) $NewValue[$ValueType] = 0;
          }
          else
          {
            if($this->Differential == 0)
            {
              if($Values[$I + 1][$ValueType] > $NewValue[$ValueType]) $NewValue[$ValueType] = $Values[$I + 1][$ValueType];
            } else {
              $Difference = $Values[$I + 1][$ValueType] - $Values[$I][$ValueType];
              if($Difference > $NewValue[$ValueType]) $NewValue[$ValueType] = $Difference;
            }
          }
        }
        else if($ValueType == 'min')
        {
          if($Values[$I + 1]['continuity'] == $this->Data['ContinuityEnabled'])
          {
            if(0 < $NewValue[$ValueType]) $NewValue[$ValueType] = 0;
          } else {
            if($this->Differential == 0)
            {
              if($Values[$I + 1][$ValueType] < $NewValue[$ValueType]) $NewValue[$ValueType] = $Values[$I + 1][$ValueType];
            } else {
              $Difference = $Values[$I + 1][$ValueType] - $Values[$I][$ValueType];
              if($Difference < $NewValue[$ValueType]) $NewValue[$ValueType] = $Difference;
            }
          }
        }
      }
      $NewValue[$ValueType] = $NewValue[$ValueType];
    }
    //if(($RightTime - $LeftTime) > 0)
    if($this->Data['Cumulative'] == 0)
    {
      $NewValue['avg'] = $NewValue['avg'] / ($RightTime - $LeftTime);
    }
    return($NewValue);
  }

  function GetTimeRange($Level)
  {
    // Get first and last time
    $Result = $this->Database->select($this->Data['DataTable'], '*', '(`measure`='.$this->Data['Id'].') AND '.
      '(`level`='.$Level.') ORDER BY `time` LIMIT 1');
    if($Result->num_rows > 0)
    {
      $Row = $Result->fetch_array();
      $AbsoluteLeftTime = MysqlDateTimeToTime($Row['time']);
    } else $AbsoluteLeftTime = 0;

    $Result = $this->Database->select($this->Data['DataTable'], '*', '(`measure`='.$this->Data['Id'].') AND '.
      '(`level`='.$Level.') ORDER BY `time` DESC LIMIT 1');
    if($Result->num_rows > 0)
    {
      $Row = $Result->fetch_array();
      $AbsoluteRightTime = MysqlDateTimeToTime($Row['time']);
    } else $AbsoluteRightTime = 0;

    return(array('left' => $AbsoluteLeftTime, 'right' => $AbsoluteRightTime));
  }

  // Load one nearest value newer then given time
  function LoadRightSideValue($Level, $Time)
  {
    $Result = array();
    $DbResult = $this->Database->select($this->Data['DataTable'], '*',
      '(`time` > "'.TimeToMysqlDateTime($Time).'") AND (`measure`='.
      $this->Data['Id'].') AND (`level`='.$Level.') ORDER BY `time` ASC LIMIT 1');
    if($DbResult->num_rows > 0)
    {
      // Found one value, simply return it
      $Row = $DbResult->fetch_array();
      $Row['time'] = MysqlDateTimeToTime($Row['time']);
      return(array($Row));
    } else
    {
      // Not found any value. Calculate it
      //$Time = $Values[count($Values) - 1]['time'] + 60;
      //array_push($Values, array('time' => $Time, 'min' => 0, 'avg' => 0, 'max' => 0, 'continuity' => 0));
      $Result[] = array('time' => ($Time + $this->TimeSegment($Level)), 'min' => 0,
        'avg' => 0, 'max' => 0, 'continuity' => 0);
      $DbResult = $this->Database->select($this->Data['DataTable'], '*',
        '(`time` < "'.TimeToMysqlDateTime($Time).'") AND (`measure`='.$this->Data['Id'].') '.
        'AND (`level`='.$Level.') ORDER BY `time` DESC LIMIT 1');
      if($DbResult->num_rows > 0)
      {
        $Row = $DbResult->fetch_array();
        array_unshift($Result, array('time' => (MysqlDateTimeToTime($Row['time']) + 10),
          'min' => 0, 'avg' => 0, 'max' => 0, 'continuity' => 0));
      }
      return($Result);
    }
  }

  // Load one nearest value older then given time
  function LoadLeftSideValue($Level, $Time)
  {
    $Result = array();
    $DbResult = $this->Database->select($this->Data['DataTable'], '*',
      '(`time` < "'.TimeToMysqlDateTime($Time).'") AND (`measure`='.$this->Data['Id'].') '.
      'AND (`level`='.$Level.') ORDER BY `time` DESC LIMIT 1');
    if($DbResult->num_rows > 0)
    {
      // Found one value, simply return it
      $Row = $DbResult->fetch_array();
      $Row['time'] = MysqlDateTimeToTime($Row['time']);
      return(array($Row));
    } else {
      // Not found any value. Calculate it
      //$Time = $Values[0]['time'] - 60;
      //array_unshift($Values, array('time' => $Time, 'min' => 0, 'avg' => 0, 'max' => 0, 'continuity' => 0));
      $Result[] = array('time' => ($Time - $this->TimeSegment($Level)), 'min' => 0,
        'avg' => 0, 'max' => 0, 'continuity' => 0);

      $DbResult = $this->Database->select($this->Data['DataTable'], '*',
        '(`time` > "'.TimeToMysqlDateTime($Time).'") AND (`measure`='.$this->Data['Id'].') '.
        'AND (`level`='.$Level.') ORDER BY `time` ASC LIMIT 1');
      if($DbResult->num_rows > 0)
      {
        $Row = $DbResult->fetch_array();
        array_push($Result, array('time' => (MysqlDateTimeToTime($Row['time']) - 10),
          'min' => 0, 'avg' => 0, 'max' => 0, 'continuity' => 0));
      }
      return($Result);
    }
  }

  function GetValues($TimeFrom, $TimeTo, $Level)
  {
  //$AbsoluteTime = GetTimeRange($this->DataId);

//  if(($TimeFrom > $AbsoluteLeftTime) and ($TimeStart < $AbsoluteRightTime) and
//    ($TimeTo > $AbsoluteLeftTime) and ($TimeTo < $AbsoluteRightTime))
//  {

    // Load values in time range
    $Result = $this->Database->select($this->Data['DataTable'], '`time`, `min`, '.
      '`avg`, `max`, `continuity`',
      '(`time` > "'.TimeToMysqlDateTime($TimeFrom).'") AND '.
      '(`time` < "'.TimeToMysqlDateTime($TimeTo).'") AND '.
      '(`measure`='.$this->Data['Id'].') AND (`level`='.$Level.') ORDER BY `time`');
    $Values = array();
    while($Row = $Result->fetch_array())
    {
      $Values[] = array('time' => MysqlDateTimeToTime($Row['time']),
        'min' => $Row['min'], 'avg' => $Row['avg'], 'max' => $Row['max'],
        'continuity' => $Row['continuity']);
    }
   // array_pop($Values);

    $Points = array();
    if(count($Values) > 0)
    {
      $Values = array_merge(
        $this->LoadLeftSideValue($Level, $TimeFrom),
        $Values,
        $this->LoadRightSideValue($Level, $TimeTo));

      $StartIndex = 0;
      $Points = array();
      for($I = 0; $I < $this->DivisionCount; $I++)
      {
        $TimeStart = $TimeFrom + (($TimeTo - $TimeFrom) / $this->DivisionCount) * $I;
        $TimeEnd = $TimeFrom + (($TimeTo - $TimeFrom) / $this->DivisionCount) * ($I + 1);

        $EndIndex = $StartIndex;
        while($Values[$EndIndex]['time'] < $TimeEnd) $EndIndex = $EndIndex + 1;
        $SubValues = array_slice($Values, $StartIndex, $EndIndex - $StartIndex + 1);
        $Points[] = $this->ComputeOneValue($TimeStart, $TimeEnd, $SubValues, $Level);
        $StartIndex = $EndIndex - 1;
      }
    } else $Points[] = array('min' => 0, 'avg' => 0, 'max' => 0);
    return($Points);
  }

  function RebuildMeasureCache()
  {
    echo('Velicina '.$this->Data['Name']."<br>\n");
    if($this->Data['Continuity'] == 0) $this->Data['ContinuityEnabled'] = 0;  // non continuous
      else $this->Data['ContinuityEnabled'] = 2;    // continuous graph

    // Clear previous items
    $DbResult = $this->Database->select($this->Data['DataTable'], 'COUNT(*)', '(`level`>0) AND (`measure`='.$this->Data['Id'].')');
    $Row = $DbResult->fetch_array();
    echo("Mazu starou cache (".$Row[0]." polozek)...");
    $this->Database->delete($this->Data['DataTable'], '(`level`>0) AND (`measure`='.$this->Data['Id'].')');
    echo("<br>\n");

    for($Level = 1; $Level <= $this->MaxLevel; $Level++)
    {
      echo('Uroven '.$Level."<br>\n");
      $TimeRange = $this->GetTimeRange($Level - 1);
      $TimeSegment = $this->TimeSegment($Level);
      $StartTime = $this->AlignTime($TimeRange['left'], $TimeSegment) - $TimeSegment;
      $EndTime = $this->AlignTime($TimeRange['right'], $TimeSegment);
      $BurstCount = 500;
      echo('For 0 to '.round(($EndTime - $StartTime) / $TimeSegment / $BurstCount)."<br>\n");
      for($I = 0; $I <= round(($EndTime - $StartTime) / $TimeSegment / $BurstCount); $I++)
      {
        echo($I.' ');
        $StartTime2 = $StartTime + $I * $BurstCount * $TimeSegment;
        $EndTime2 = $StartTime + ($I + 1) * $BurstCount * $TimeSegment;
        $Values = array();
        $DbResult = $this->Database->select($this->Data['DataTable'], '*', '(`time` > "'.TimeToMysqlDateTime($StartTime2).'") AND (`time` < "'.
          TimeToMysqlDateTime($EndTime2).'") AND (`measure`='.$this->Data['Id'].') AND (`level`='.($Level - 1).') ORDER BY `time`');
        while($Row = $DbResult->fetch_array())
        {
          $Row['time'] = MysqlDateTimeToTime($Row['time']);
          $Values[] = $Row;
        }

        if(count($Values) > 0)
        {
          $Values = array_merge(
            $this->LoadLeftSideValue($Level - 1, $StartTime2),
            $Values,
            $this->LoadRightSideValue($Level - 1, $EndTime2));

          $StartIndex = 0;
          for($B = 0; $B < $BurstCount; $B++)
          {
            echo('.');
            $StartTime3 = $StartTime2 + (($EndTime2 - $StartTime2) / $BurstCount) * $B;
            $EndTime3 = $StartTime2 + (($EndTime2 - $StartTime2) / $BurstCount) * ($B + 1);

            $EndIndex = $StartIndex;
            while($Values[$EndIndex]['time'] < $EndTime3) $EndIndex = $EndIndex + 1;
            $SubValues = array_slice($Values, $StartIndex, $EndIndex - $StartIndex + 1);
            if(count($SubValues) > 2)
            {
              $Point = $this->ComputeOneValue($StartTime3, $EndTime3, $SubValues, $Level);
              $Continuity = $SubValues[1]['continuity'];
              $this->Database->insert($this->Data['DataTable'], array('level' => $Level, 'measure' => $this->Data['Id'],
                'min' => $Point['min'], 'avg' => $Point['avg'], 'max' => $Point['max'],
                'continuity' => $Continuity, 'time' => TimeToMysqlDateTime($StartTime3 + ($EndTime3 - $StartTime3) / 2)));
            }
            $StartIndex = $EndIndex - 1;
          }
        }
        // Load values in time range
        //array_pop($NextValues);
      }
      echo("Uroven dokoncena<br>\n");
      $DbResult = $this->Database->select($this->Data['DataTable'], 'COUNT(*)',
        '(`level`='.$Level.') AND (`measure`='.$this->Data['Id'].')');
      $Row = $DbResult->fetch_array();
      echo("Vlozeno ".$Row[0]." polozek.<br>\n");
    }
  }

  function RebuildAllMeasuresCache()
  {
    $Result = $this->Database->select('measure', 'Id');
    while($Row = $Result->fetch_array())
    {
      $Measure = new Measure();
      $Measure->Load($Row['Id']);
      $Measure->RebuildMeasureCache();
      echo('Velicina id '.$Row['Id'].' dokoncena<br/>');
    }
  }

  function InitMeasureDataTable()
  {
    $this->Database->query('CREATE TABLE `data_'.$this->Data['Name'].'` (
      `Time` TIMESTAMP NOT NULL ,
      `Avg` '.$this->Data['DataType'].' NOT NULL ,
      `Continuity` BOOL NOT NULL
      ) ENGINE = MYISAM ;');

    $this->Database->query('CREATE TABLE `data_'.$this->Data['Name'].'_cache` (
`Time` TIMESTAMP NOT NULL ,
`Level` TINYINT NOT NULL ,
`Min` '.$this->Data['DataType'].' NOT NULL ,
`Avg` '.$this->Data['DataType'].' NOT NULL ,
`Max` '.$this->Data['DataType'].' NOT NULL ,
`Continuity` BOOL NOT NULL
) ENGINE = MYISAM ;');
  }
}
